#
#  Copyright 1999 Illinois Institute of Technology
#
#  See the file "LICENSE" for information on usage and redistribution
#  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#

proc create_experiment_window { type record_list } {

	set toplevel .experiment_window

	set status [ catch { toplevel $toplevel } result ]

	# If the window already exists, bring it to the top.

	if { $status != 0 } {
		raise $toplevel
		focus $toplevel

		select_experiment_type $toplevel $type
		return
	}
	focus $toplevel

	wm title $toplevel "Spectroscopy & MAD Data Collection Setup"

	create_experiment_menubar $toplevel $record_list

	create_experiment_selector $toplevel

	frame $toplevel.sep0 -height 2 -borderwidth 1 -relief sunken
	pack $toplevel.sep0 -side top -expand yes -fill x -pady 5

	create_main_spectroscopy_window $toplevel $record_list

	frame $toplevel.sep1 -height 2 -borderwidth 1 -relief sunken
	pack $toplevel.sep1 -side top -expand yes -fill x -pady 5

	create_mad_bottom_window $toplevel $record_list

	select_experiment_type $toplevel $type
}

proc create_experiment_menubar { toplevel record_list } {

	set menubar $toplevel.menubar

	menu $menubar
	$toplevel configure -menu $menubar

	menu $menubar.help -tearoff 0
	$menubar add cascade -label Help -menu $menubar.help

	$menubar.help add command -label "Help for Spectroscopy" \
			-command [ code display_help "spectroscopy.html" ]
	$menubar.help add command -label "Help for MAD" \
			-command [ code display_help "mad.html" ]
	$menubar.help add command -label "Help for Chooch" \
			-command [ code display_help "chooch.html" ]
}

proc create_experiment_selector { toplevel } {

	iwidgets::optionmenu $toplevel.type -labeltext "Experiment Type" \
		-command [ code select_experiment_command $toplevel ]

	pack $toplevel.type

	$toplevel.type insert end Spectroscopy MAD
	$toplevel.type select MAD
}

proc select_experiment_command { toplevel } {

	set type [ $toplevel.type get ]

	select_experiment_type $toplevel $type
}

proc select_experiment_type { toplevel type } {

	switch -exact -- $type {
		Spectroscopy { pack forget $toplevel.bottom; \
			pack forget $toplevel.sep2 }

		MAD { pack $toplevel.bottom -side top -after $toplevel.sep1; \
			pack $toplevel.sep2 -side top -expand yes \
				-fill x -pady 5 -after $toplevel.bottom }

		default { tk_dialog .unexpected_experiment_type \
			"Unexpected Experiment Type" \
	"Experiment type should be 'Spectroscopy' or 'MAD', not '$type'" \
			error 0 "OK" }
	}

	$toplevel.type select $type
}

