#
#  Copyright 1999-2002 Illinois Institute of Technology
#
#  See the file "LICENSE" for information on usage and redistribution
#  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#

proc setup_imcagui_database { motor_database scan_database } {

	#
	# Create a new record list.
	#

	set record_list [ Mx::RecordList ::mxrecordlist_#auto ]

	#
	# Enable plotting.
	#

	$record_list set_plot_enable 1

	#
	# Read the motor and scan databases.
	#

	$record_list read_database $motor_database
	$record_list read_database $scan_database

	#
	# Finish initializing MX.
	#

	$record_list finish_database_initialization
	$record_list initialize_hardware 0

	return $record_list
}

proc create_imcagui_gui { record_list } {

	create_imcagui_menubar $record_list

	create_imcagui_main_window $record_list

	setup_imcagui_callbacks $record_list
}

proc find_callback_record { record_list item_name record_name } {

	global callback_records

	set status [ catch { $record_list get_record $record_name } \
			record ]

	if { $status != 0 } {
		log_puts \
		"Error: '$record_name' record not found in database."

		return -code error
	}

	set callback_records($item_name) $record
}

proc add_imcagui_callback { callback } {
	global callback_list

	lappend callback_list $callback
}

proc delete_imcagui_callback { callback } {
	global callback_list

	set i [ lsearch -exact $callback_list $callback ]

	if { $i < 0 } {
		set t \
	    "Attempt to delete callback '$callback' from callback list failed."

		tk_dialog .delete_callback_failed \
		"Callback Deletion Failed." $t error 0 "OK"

		return
	}

	set callback_list [ lreplace $callback_list $i $i ]
}

proc start_imcagui_callbacks { } {
	set callback_interval 1000	;# in milliseconds

	set callback_seconds [ expr ( 0.001 * double( $callback_interval ) ) ]

	set elapsed_time_record $IMCAGui::elapsed_time_record

	set current_time [ $elapsed_time_record get_position ]

	set IMCAGui::previous_callback_time \
			[ expr ( $current_time - $callback_seconds ) ]

	set IMCAGui::callback_interval $callback_interval

	after 1 [ list process_imcagui_callbacks ]
}

proc process_imcagui_callbacks { } {

	global callback_list
	global callback_records
	global callback_values

#	set elapsed_milliseconds [ elapsed_milliseconds ]
#
#	log_puts "Before process = $elapsed_milliseconds ms"

	foreach name [ array names callback_records ] {
	    set record $callback_records($name)

	    set superclass [ $record get_field "mx_superclass" ]

	    if { $superclass == "variable" } {
		if [ catch { $record receive_variable } \
						callback_values($name) ] \
		{
			set callback_values($name) "*****"
		}
	    } elseif { $superclass == "device" } {

		set class [ $record get_field "mx_class" ]

		if { $class == "motor" } {
		    if [ catch { $record get_position } \
						callback_values($name) ] \
		    {
			set callback_values($name) "*****"
		    }

		    set busyname "${name}_busy"

		    if [ catch { $record is_busy } \
						callback_values($busyname) ] \
		    {
			set callback_values($busyname) "*****"
		    }
		} elseif { $class == "digital_input" } {
		    if [ catch { $record read } callback_values($name) ] {
			set callback_values($name) "*****"
		    }
		} elseif { $class == "relay" } {
		    if [ catch { $record get_relay_status } \
						callback_values($name) ] {
			set callback_values($name) "*****"
		    }
		} else {
		    set callback_values($name) "+++++"
		}
	    } else {
		set callback_values($name) "+++++"
	    }

	    #
	    # process_imcagui_callbacks can take a very long time ( ~1 sec )
	    # to execute.  Thus, it is vitally important that the 'update'
	    # command be invoked periodically so that the GUI seems more
	    # responsive to the user.
	    #
	    # Apparently, 'update idletasks' is not sufficient for this
	    # since imcagui never becomes idle during the time that 
	    # process_imcagui_callbacks is running.
	    #

	    update
	}

	if [ catch { energy_from_lambda $callback_values(lambda) } \
						callback_values(energy) ] {
		set callback_values(energy) "!!!!!"
	}

	foreach callback $callback_list {
		$callback
	}

#	set elapsed_milliseconds [ elapsed_milliseconds ]
#
#	log_puts "Process time = $elapsed_milliseconds ms"

	after $IMCAGui::callback_interval [ list process_imcagui_callbacks ]
}

proc elapsed_milliseconds { } {
	set elapsed_time_record $IMCAGui::elapsed_time_record

	if { $elapsed_time_record == "" } {
		return -1
	}

	set previous_time $IMCAGui::previous_callback_time

	set current_time [ $elapsed_time_record get_position ]

	set IMCAGui::previous_callback_time $current_time

	set elapsed_milliseconds \
	    [ expr ( round( 1000.0 * ( $current_time - $previous_time ) ) ) ]

	if { $elapsed_milliseconds < 0 } {
		log_puts 
		set t    "Warning: elapsed milliseconds since "
		append t "last callback = $elapsed_milliseconds "
		append t "is less than zero."
	}

	return $elapsed_milliseconds
}

proc create_imcagui_menubar { record_list } {
	menu .menubar
	. configure -menu .menubar

	###############################################################
	#
	#   Add the File Menu
	#

	menu .menubar.file -tearoff 0
	.menubar add cascade -label File -menu .menubar.file

	.menubar.file add command -label Print -command [ code print_window . ]

	.menubar.file add command -label Staff \
		-command [ code eval exec "mxgui &" ]

	.menubar.file add command -label Exit -command exit

	###############################################################
	#
	#   Add the Edit Menu
	#

	menu .menubar.edit -tearoff 0
	.menubar add cascade -label Edit -menu .menubar.edit

	.menubar.edit add command -label "Set Default Printer" \
				-command set_default_printer

	###############################################################
	#
	#   Add the Variables Menu
	#

	menu .menubar.variables -tearoff 0
	.menubar add cascade -label Variables -menu .menubar.variables

	.menubar.variables add command -label "Edge Energy" \
				-command set_energy

	if { $IMCAGui::have_undulator } {
		.menubar.variables add command -label "ID_ev_enabled" \
				-command set_ev_enabled

		.menubar.variables add command -label "ID_ev_harmonic" \
				-command set_harmonic

		.menubar.variables add command -label "ID_ev_params" \
				-command set_ev_params
	}
	###############################################################
	#
	#   Add the Spectroscopy / MAD Menu
	#

	menu .menubar.mad -tearoff 0
	.menubar add cascade -label Experiment -menu .menubar.mad

	.menubar.mad add command -label Spectroscopy -command [ code \
		create_experiment_window Spectroscopy $record_list ]

	.menubar.mad add command -label MAD -command [ code \
		create_experiment_window MAD $record_list ]

	###############################################################
	#
	#   Add the Help Menu
	#

	create_help_menu .menubar
}

#
#  Create the main window.
#
proc create_imcagui_main_window { record_list } {

	set window .main

	frame .main
	pack .main -before .message_log

	#-----

	set frame .main.energy_frame

	frame $frame
	pack $frame -side top

	frame $frame.grid
	pack $frame.grid -side left

	create_imcagui_energy_frame $record_list $frame.grid

	frame $frame.sep
	pack $frame.sep -side left -padx 5

	button $frame.move_button -text Move -width 6 -command \
			[ code move_lambda $frame ]

	pack $frame.move_button -side left

	set IMCAGui::default_button_background \
			[ $frame.move_button cget -background ]

	$frame.move_button configure -disabledforeground \
			[ $frame.move_button cget -foreground ]

	#-----

	frame .main.sep1 -height 2 -borderwidth 1 -relief sunken
	pack .main.sep1 -expand yes -fill x -pady 5 -side top

	#-----

	set frame .main.harmonic_frame

	frame $frame
	pack $frame -side top

	create_harmonic_frame $frame

	#-----

	frame .main.sep2 -height 2 -borderwidth 1 -relief sunken
	pack .main.sep2 -expand yes -fill x -pady 5 -side top

	#-----

	set slit_frame .main.slit_frame

	frame $slit_frame
	pack $slit_frame -side top

	create_slit_frame $slit_frame

	if { $IMCAGui::have_filters } {
		#-----

		frame .main.sep3 -height 2 -borderwidth 1 -relief sunken
		pack .main.sep3 -expand yes -fill x -pady 5 -side top

		#-----

	        set frame .main.filter_frame

	        frame $frame
	        pack $frame -side top

		create_filter_frame $frame $record_list
		frame $frame.sep
		pack $frame.sep -side left -padx 5
	}

	#-----

	frame .main.sep5 -height 2 -borderwidth 1 -relief sunken
	pack .main.sep5 -expand yes -fill x -pady 5 -side top

	set frame .main.bottom

	frame $frame
	pack $frame -side top

	menubutton $frame.optimize -text "Optimize Intensity..." \
		-background limegreen -relief raised -padx 5 -pady 5

	$frame.optimize configure -activebackground limegreen

	set menu $frame.optimize.menu

	menu $menu -tearoff 0 -postcommand \
				[ code optimize_menu_handler $frame ]

	$frame.optimize configure -menu $menu

	pack $frame.optimize -side left

	#-----

	frame .main.sep_last -height 2 -borderwidth 1 -relief sunken
	pack .main.sep_last -expand yes -fill x -pady 5 -side top

	#-----

	add_imcagui_callback main_window_callback
}

proc main_window_callback { } {

	global callback_values

	set window .main

	set frame $window.energy_frame

	$frame.grid.energy_value configure -text \
		[ imcagui_format "%10.3f" $callback_values(energy) ]

	$frame.grid.lambda_value configure -text \
		[ imcagui_format "%8.4f"  $callback_values(lambda) ]

	set busy_list [ list ]
	lappend busy_list $callback_values(lambda_busy)

	compute_move_state $frame.move_button $busy_list "Move"

	if { $IMCAGui::have_undulator } {

		$frame.grid.id_value configure -text \
			[ imcagui_format "%10.3f" $callback_values(id_ev) ]
	}

	set frame $window.harmonic_frame

	if { $IMCAGui::have_undulator } {
		$frame.harmonic_value configure -text \
		    [ imcagui_format "%2d" $callback_values(id_ev_harmonic) ]
	}

	$frame.d_spacing_value configure -text \
		[ imcagui_format "%8.5f" $callback_values(d_spacing) ]

	set frame $window.slit_frame

	$frame.horiz_size_value configure -text \
		[ imcagui_format "%10.3f" $callback_values(hslit_size) ]

	$frame.vert_size_value configure -text \
		[ imcagui_format "%10.3f" $callback_values(vslit_size) ]

	if { $IMCAGui::show_slit_positions } {
		$frame.horiz_center_value configure -text \
		    [ imcagui_format "%10.3f" $callback_values(hslit_center) ]

		$frame.vert_center_value configure -text \
		    [ imcagui_format "%10.3f" $callback_values(vslit_center) ]
	}

	set busy_list [ list ]
	lappend busy_list $callback_values(hslit_size_busy)
	lappend busy_list $callback_values(vslit_size_busy)

	compute_move_state $frame.move_sizes $busy_list "Change Sizes"

	if { $IMCAGui::show_slit_positions } {
		set busy_list [ list ]
		lappend busy_list $callback_values(hslit_center_busy)
		lappend busy_list $callback_values(vslit_center_busy)

		compute_move_state $frame.move_centers \
					$busy_list "Change Positions"
	}

	if { $IMCAGui::have_filters } {
		update_filter_frame .main.filter_frame
        }

}

proc imcagui_format { format_specifier value } {

	if [ catch { format $format_specifier $value } result ] {
		set result $value
	}
	return $result
}

proc set_move_state_to_moving { button } {
	$button configure -text "Moving"
	$button configure -background yellow
	$button configure -activebackground yellow
	$button configure -state disabled

	update
}

proc compute_move_state { button busy_list default } {

	set fault 0
	set busy 0

	foreach item $busy_list {
		if { $item < 0 } {
			set fault 1
		}
		if { $item > 0 } {
			set busy 1
		}
	}

	if { $fault } {
		$button configure -text "Fault"
		$button configure -background red
		$button configure -activebackground red
		$button configure -state disabled
	} else {
		if { $busy } {
			$button configure -text "Moving"
			$button configure -background yellow
			$button configure -activebackground yellow
			$button configure -state disabled
		} else {
			set bg $IMCAGui::default_button_background

			$button configure -text $default
			$button configure -background limegreen
			$button configure -activebackground limegreen
			$button configure -state normal
		}
	}
	update
}

############################################################################
#
#  Create the energy and wavelength display frame.
#
proc create_imcagui_energy_frame { record_list frame } {

	set row 0

	label $frame.current_size_label -text Current -width 10
	label $frame.desired_size_label -text Desired 

	grid $frame.current_size_label -row $row -column 1
	grid $frame.desired_size_label -row $row -column 3

	incr row

	label $frame.energy_label -text Energy
	label $frame.energy_value -text "---" -fg blue
	label $frame.energy_units1 -text "eV   "
	label $frame.energy_units2 -text eV

	iwidgets::entryfield $frame.energy_desired -width 10 \
		-validate { iwidgets::Entryfield::real %P } \
		-command [ code energy_desired_update $frame ]

	grid $frame.energy_label   -row $row -column 0 -sticky w
	grid $frame.energy_value   -row $row -column 1 -sticky e
	grid $frame.energy_units1  -row $row -column 2 -sticky w
	grid $frame.energy_desired -row $row -column 3
	grid $frame.energy_units2  -row $row -column 4 -sticky w

	incr row

	label $frame.lambda_label -text Wavelength
	label $frame.lambda_value -text "---" -fg blue
	label $frame.lambda_units1 -text A
	label $frame.lambda_units2 -text A

	iwidgets::entryfield $frame.lambda_desired -width 10 \
		-validate { iwidgets::Entryfield::real %P } \
		-command [ code lambda_desired_update $frame ]

	grid $frame.lambda_label   -row $row -column 0 -sticky w
	grid $frame.lambda_value   -row $row -column 1 -sticky e
	grid $frame.lambda_units1  -row $row -column 2 -sticky w
	grid $frame.lambda_desired -row $row -column 3
	grid $frame.lambda_units2  -row $row -column 4 -sticky w

	if { $IMCAGui::have_undulator } {
		incr row

		label $frame.id_label -text "Gap Energy"
		label $frame.id_value -text "---" -fg blue
		label $frame.id_units -text eV

		grid $frame.id_label -row $row -column 0 -sticky w
		grid $frame.id_value -row $row -column 1 -sticky e
		grid $frame.id_units -row $row -column 2 -sticky w
	}

	bind $frame.energy_desired.lwchildsite.entry <KeyRelease> \
		[ code energy_desired_update $frame ]

	bind $frame.lambda_desired.lwchildsite.entry <KeyRelease> \
		[ code lambda_desired_update $frame ]
}

proc energy_desired_update { frame } {

	set energy [ $frame.energy_desired get ]

	set lambda [ lambda_from_energy $energy ]

	$frame.lambda_desired clear
	$frame.lambda_desired insert end $lambda
}

proc lambda_desired_update { frame } {

	set lambda [ $frame.lambda_desired get ]

	set energy [ energy_from_lambda $lambda ]

	$frame.energy_desired clear
	$frame.energy_desired insert end $energy
}

proc move_lambda { frame } {

	global callback_records
	global callback_values

	set grid $frame.grid

	set lambda_requested [ $grid.lambda_desired get ]

	if { $lambda_requested == "" } {

		set energy_requested [ $grid.energy_desired get ]

		if { $energy_requested == "" } {
			set t "No energy or wavelength value requested."

			tk_dialog .no_wavelength_dialog \
				"No Value Specified" $t error 0 "OK"

			return
		}

		set lambda_requested \
			[ lambda_from_energy $energy_requested ]
	}

	set status [ catch { expr $lambda_requested } junk ]

	if { $status != 0 } {
		set t "Invalid wavelength value '$lambda_requested' requested."

		tk_dialog .invalid_wavelength_dialog \
			"Invalid Wavelength" $t error 0 "OK"

		return
	}

#	if [ expr { $lambda_requested > 1.16 } ] {
#		$callback_records(id_ev_harmonic) send_variable 1
#	} else {
#		$callback_records(id_ev_harmonic) send_variable 3
#	}

	set_move_state_to_moving $frame.move_button

	$callback_records(lambda) move_absolute $lambda_requested
}

############################################################################

proc create_harmonic_frame { frame } {

	if { $IMCAGui::have_undulator } {
		label $frame.harmonic_label -text "Undulator\nHarmonic" 
		pack $frame.harmonic_label -side left

		frame $frame.sep1 -width 1 -borderwidth 1
		pack $frame.sep1 -side left -expand yes -fill y -padx 5

		label $frame.harmonic_value -text "-" -fg blue
		pack $frame.harmonic_value -side left

		frame $frame.sep2 -width 1 -borderwidth 1
		pack $frame.sep2 -side left -expand yes -fill y -padx 20
	}

	label $frame.d_spacing_label -text "Monochromator\nd Spacing"
	pack $frame.d_spacing_label -side left

	frame $frame.sep3 -width 1 -borderwidth 1
	pack $frame.sep3 -side left -expand yes -fill y -padx 5

	label $frame.d_spacing_value -text "---" -fg blue
	pack $frame.d_spacing_value -side left

	frame $frame.sep4 -width 1 -borderwidth 1
	pack $frame.sep4 -side left -expand yes -fill y -padx 5

	label $frame.d_spacing_units -text "A"
	pack $frame.d_spacing_units -side left
}

############################################################################

proc create_slit_frame { frame } {

	set row 0

	label $frame.frame_label -text "Slits:"

	label $frame.size_label -text "Slit Sizes"

	grid $frame.frame_label  -row $row -column 0
	grid $frame.size_label   -row $row -column 1 -columnspan 2

	if { $IMCAGui::show_slit_positions } {
		label $frame.center_label -text "Slit Positions"
		grid $frame.center_label -row $row -column 3 -columnspan 2
	}

	incr row

	label $frame.current_size_label -text Current -width 10
	label $frame.desired_size_label -text Desired

	grid $frame.current_size_label   -row $row -column 1
	grid $frame.desired_size_label   -row $row -column 2

	if { $IMCAGui::show_slit_positions } {
		label $frame.current_center_label -text Current -width 10
		label $frame.desired_center_label -text Desired
		grid $frame.current_center_label -row $row -column 3
		grid $frame.desired_center_label -row $row -column 4
	}

	incr row

	label $frame.vert_label -text "Vertical"
	label $frame.vert_size_value -text "---" -fg blue

	iwidgets::entryfield $frame.vert_size_desired -width 8 \
		-validate { iwidgets::Entryfield::real %P }

	grid $frame.vert_label          -row $row -column 0 -sticky w
	grid $frame.vert_size_value     -row $row -column 1 -sticky e
	grid $frame.vert_size_desired   -row $row -column 2

	if { $IMCAGui::show_slit_positions } {
		label $frame.vert_center_value -text "---" -fg blue

		iwidgets::entryfield $frame.vert_center_desired -width 8 \
			-validate { iwidgets::Entryfield::real %P }

		grid $frame.vert_center_value   -row $row -column 3 -sticky e
		grid $frame.vert_center_desired -row $row -column 4
	}

	label $frame.vert_units -text "um"

	grid $frame.vert_units          -row $row -column 5 -sticky w

	incr row

	label $frame.horiz_label -text "Horizontal"
	label $frame.horiz_size_value -text "---" -fg blue

	iwidgets::entryfield $frame.horiz_size_desired -width 8 \
		-validate { iwidgets::Entryfield::real %P }

	grid $frame.horiz_label          -row $row -column 0 -sticky w
	grid $frame.horiz_size_value     -row $row -column 1 -sticky e
	grid $frame.horiz_size_desired   -row $row -column 2

	if { $IMCAGui::show_slit_positions } {
		label $frame.horiz_center_value -text "---" -fg blue

		iwidgets::entryfield $frame.horiz_center_desired -width 8 \
			-validate { iwidgets::Entryfield::real %P }

		grid $frame.horiz_center_value   -row $row -column 3 -sticky e
		grid $frame.horiz_center_desired -row $row -column 4
	}

	label $frame.horiz_units -text "um"

	grid $frame.horiz_units          -row $row -column 5 -sticky w

	incr row

	button $frame.move_sizes -text "Change Sizes" \
			-command [ code move_slit_sizes $frame ]

	$frame.move_sizes configure -disabledforeground \
			[ $frame.move_sizes cget -foreground ]

	grid $frame.move_sizes -pady 5 -row $row -column 1 -columnspan 2

	if { $IMCAGui::show_slit_positions } {
		button $frame.move_centers -text "Change Positions" \
			-command [ code move_slit_centers $frame ]

		$frame.move_centers configure -disabledforeground \
			[ $frame.move_centers cget -foreground ]

		grid $frame.move_centers \
			-pady 5 -row $row -column 3 -columnspan 2
	}
}

proc move_slit_sizes { frame } {

	global callback_records

	set_move_state_to_moving $frame.move_sizes

	set horiz_requested [ $frame.horiz_size_desired get ]

	if { $horiz_requested != "" } {
		$callback_records(hslit_size) move_absolute $horiz_requested
	}

	set vert_requested [ $frame.vert_size_desired get ]

	if { $vert_requested != "" } {
		$callback_records(vslit_size) move_absolute $vert_requested
	}
}

proc move_slit_centers { frame } {

	global callback_records

	set_move_state_to_moving $frame.move_centers

	set horiz_requested [ $frame.horiz_center_desired get ]

	if { $horiz_requested != "" } {
		$callback_records(hslit_center) move_absolute $horiz_requested
	}

	set vert_requested [ $frame.vert_center_desired get ]

	if { $vert_requested != "" } {
		$callback_records(vslit_center) move_absolute $vert_requested
	}
}

proc set_energy {} {
        global callback_records

        set toplevel .energy_window

        set status [ catch { toplevel $toplevel } ]

        # If the window already exists, bring it to the top.

        if { $status != 0 } {
                raise $toplevel
                focus $toplevel
                return
        }

        wm title $toplevel "Edge Energy Control"

        set grid $toplevel.grid

        frame $grid
        pack $grid -side top
# set up entry field with current value
	iwidgets::entryfield $grid.value -labeltext "Edge energy"
	grid $grid.value     -row 0 -column 0
        window_callback edge_energy $grid.value
# set up bottom row
        set bottom $toplevel.bottom

        frame $bottom
        pack $bottom -side top

        button $bottom.move -text "Set" \
                -command [ code set_value \
                edge_energy $grid ]

        pack $bottom.move -side left

        frame $bottom.sep
        pack $bottom.sep -side left -padx 5

        button $bottom.dismiss -text "Dismiss" \
                -command [ code destroy $toplevel ]
        pack $bottom.dismiss -side left

}

proc set_ev_enabled {} {
        global callback_records

        set toplevel .ev_enable_window

        set status [ catch { toplevel $toplevel } ]

        # If the window already exists, bring it to the top.

        if { $status != 0 } {
                raise $toplevel
                focus $toplevel
                return
        }

        wm title $toplevel "ID Gap Enable Control"

        set grid $toplevel.grid

        frame $grid
        pack $grid -side top
# set up entry field with current value
	iwidgets::entryfield $grid.value -labeltext "Gap Enable"
	grid $grid.value     -row 0 -column 0
        window_callback id_ev_enabled $grid.value
# set up bottom row
        set bottom $toplevel.bottom

        frame $bottom
        pack $bottom -side top

        button $bottom.move -text "Set" \
                -command [ code set_value \
                id_ev_enabled $grid ]

        pack $bottom.move -side left

        frame $bottom.sep
        pack $bottom.sep -side left -padx 5

        button $bottom.dismiss -text "Dismiss" \
                -command [ code destroy $toplevel ]
        pack $bottom.dismiss -side left
}

proc set_ev_params {} {

        global callback_records

        set toplevel .ev_params_window

        set status [ catch { toplevel $toplevel } ]

        # If the window already exists, bring it to the top.

        if { $status != 0 } {
                raise $toplevel
                focus $toplevel
                return
        }

        wm title $toplevel "Gap Energy Offset Parameters"

        set grid $toplevel.grid

        frame $grid
        pack $grid -side top
# set up entry field with current value
	iwidgets::entryfield $grid.value -labeltext "Gap Parameters"
	grid $grid.value     -row 0 -column 0
        window_callback id_ev_params $grid.value
# set up bottom row
        set bottom $toplevel.bottom

        frame $bottom
        pack $bottom -side top

        button $bottom.move -text "Set" \
                -command [ code set_value \
                id_ev_params $grid ]

        pack $bottom.move -side left

        frame $bottom.sep
        pack $bottom.sep -side left -padx 5

        button $bottom.dismiss -text "Dismiss" \
                -command [ code destroy $toplevel ]
        pack $bottom.dismiss -side left
}

proc set_harmonic {} {
        global callback_records

        set toplevel .harmonic_window

        set status [ catch { toplevel $toplevel } ]

        # If the window already exists, bring it to the top.

        if { $status != 0 } {
                raise $toplevel
                focus $toplevel
                return
        }

        wm title $toplevel "Harmonic Control"

        set grid $toplevel.grid

        frame $grid
        pack $grid -side top
# set up entry field with current value
	iwidgets::entryfield $grid.value -labeltext "Harmonic"
	grid $grid.value     -row 0 -column 0
        window_callback id_ev_harmonic $grid.value
# set up bottom row
        set bottom $toplevel.bottom

        frame $bottom
        pack $bottom -side top

        button $bottom.move -text "Set" \
                -command [ code set_value \
                id_ev_harmonic $grid ]

        pack $bottom.move -side left

        frame $bottom.sep
        pack $bottom.sep -side left -padx 5

        button $bottom.dismiss -text "Dismiss" \
                -command [ code destroy $toplevel ]

        pack $bottom.dismiss -side left


}
proc window_callback { record window } {
        global callback_records
	global callback_values
#        set window .harmonic_window.grid.value

#        if { [ winfo exists $window ] == 0 } {
#                delete_imcagui_callback harmonic_window_callback
#                return
#        }
# get value of variable
        if [ catch { $callback_records($record) \
                                receive_variable } \
                                callback_values($record) ] \
        {
                set callback_values($record) "*****"
        }

#        { $callback_records(id_ev_harmonic) receive_variable } $value
# set edit field text
        $window clear
        $window insert end [ imcagui_format "%2d" \
			 $callback_values($record) ]

}
proc set_value { record grid } {
        global callback_records
	global callback_values
    set value_desired [ $grid.value get ]
    $callback_records($record) send_variable $value_desired
}
